#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <xcb/xcb.h>

#define ARRAY_SIZE(a) (sizeof(a) / sizeof((a)[0]))
#define FONT "-*-terminus-medium-r-*-*-16-*-*-*-*-*-*-*"

static xcb_connection_t *conn;
static xcb_window_t win;
static xcb_gcontext_t foreground;
static uint32_t foreground_color = 9398012;
static uint32_t background_color = 0;
static xcb_font_t font;

static struct {
	const char *name;
	int nr;
} event_counters[] = {
	[XCB_KEY_PRESS] = { "key_press", 0 },
	[XCB_KEY_RELEASE] = { "key_release", 0 },
	[XCB_BUTTON_PRESS] = { "button_press", 0 },
	[XCB_BUTTON_RELEASE] = { "button_release", 0 },
	[XCB_MOTION_NOTIFY] = { "motion_notify", 0 },
	[XCB_ENTER_NOTIFY] = { "enter_notify", 0 },
	[XCB_LEAVE_NOTIFY] = { "leave_notify", 0 },
	[XCB_FOCUS_IN] = { "focus_in", 0 },
	[XCB_FOCUS_OUT] = { "focus_out", 0 },
	[XCB_KEYMAP_NOTIFY] = { "keymap_notify", 0 },
	[XCB_EXPOSE] = { "expose", 0 },
	[XCB_GRAPHICS_EXPOSURE] = { "graphics_exposure", 0 },
	[XCB_NO_EXPOSURE] = { "no_exposure", 0 },
	[XCB_VISIBILITY_NOTIFY] = { "visibility_notify", 0 },
	[XCB_CREATE_NOTIFY] = { "create_notify", 0 },
	[XCB_DESTROY_NOTIFY] = { "destroy_notify", 0 },
	[XCB_UNMAP_NOTIFY] = { "unmap_notify", 0 },
	[XCB_MAP_NOTIFY] = { "map_notify", 0 },
	[XCB_MAP_REQUEST] = { "map_request", 0 },
	[XCB_REPARENT_NOTIFY] = { "reparent_notify", 0 },
	[XCB_CONFIGURE_NOTIFY] = { "configure_notify", 0 },
	[XCB_CONFIGURE_REQUEST] = { "configure_request", 0 },
	[XCB_GRAVITY_NOTIFY] = { "gravity_notify", 0 },
	[XCB_RESIZE_REQUEST] = { "resize_request", 0 },
	[XCB_CIRCULATE_NOTIFY] = { "circulate_notify", 0 },
	[XCB_CIRCULATE_REQUEST] = { "circulate_request", 0 },
	[XCB_PROPERTY_NOTIFY] = { "property_notify", 0 },
	[XCB_SELECTION_CLEAR] = { "selection_clear", 0 },
	[XCB_SELECTION_REQUEST] = { "selection_request", 0 },
	[XCB_SELECTION_NOTIFY] = { "selection_notify", 0 },
	[XCB_COLORMAP_NOTIFY] = { "colormap_notify", 0 },
	[XCB_CLIENT_MESSAGE] = { "client_message", 0 },
	[XCB_MAPPING_NOTIFY] = { "mapping_notify", 0 },
};

static void
panic(const char *m)
{
	fprintf(stderr, "%s\n", m);
	exit(1);
}

static const char *
connection_error_message(int err)
{
	switch (err) {
	case XCB_CONN_ERROR:			return "IO error";
	case XCB_CONN_CLOSED_EXT_NOTSUPPORTED:	return "extension not supported";
	case XCB_CONN_CLOSED_MEM_INSUFFICIENT:	return "insufficient memory";
	case XCB_CONN_CLOSED_REQ_LEN_EXCEED:	return "request length exceeded";
	case XCB_CONN_CLOSED_PARSE_ERR:		return "parse error";
	}
	return "unknown";
}

static void
open_connection(void)
{
	int err;
	conn = xcb_connect(NULL, NULL);
	err = xcb_connection_has_error(conn);
	if (err)
		panic(connection_error_message(err));
}

static void
open_font(void)
{
	font = xcb_generate_id(conn);
	xcb_open_font(conn, font, sizeof(FONT)-1, FONT);
}

static void
create_foreground(void)
{
	uint32_t mask =
		XCB_GC_FOREGROUND | XCB_GC_BACKGROUND | XCB_GC_FONT |
		XCB_GC_GRAPHICS_EXPOSURES;
	uint32_t values[] = {
		foreground_color,
		background_color,
		font,
		0,
	};
	foreground = xcb_generate_id(conn);
	xcb_create_gc(conn, foreground, win, mask, values);
}

#define ALL_EVENTS \
    XCB_EVENT_MASK_KEY_PRESS \
    | XCB_EVENT_MASK_KEY_RELEASE \
    | XCB_EVENT_MASK_BUTTON_PRESS \
    | XCB_EVENT_MASK_BUTTON_RELEASE \
    | XCB_EVENT_MASK_ENTER_WINDOW \
    | XCB_EVENT_MASK_LEAVE_WINDOW \
    | XCB_EVENT_MASK_POINTER_MOTION \
    | XCB_EVENT_MASK_POINTER_MOTION_HINT \
    | XCB_EVENT_MASK_BUTTON_1_MOTION \
    | XCB_EVENT_MASK_BUTTON_2_MOTION \
    | XCB_EVENT_MASK_BUTTON_3_MOTION \
    | XCB_EVENT_MASK_BUTTON_4_MOTION \
    | XCB_EVENT_MASK_BUTTON_5_MOTION \
    | XCB_EVENT_MASK_BUTTON_MOTION \
    | XCB_EVENT_MASK_KEYMAP_STATE \
    | XCB_EVENT_MASK_EXPOSURE \
    | XCB_EVENT_MASK_VISIBILITY_CHANGE \
    | XCB_EVENT_MASK_STRUCTURE_NOTIFY \
    | XCB_EVENT_MASK_RESIZE_REDIRECT \
    | XCB_EVENT_MASK_SUBSTRUCTURE_NOTIFY \
    | XCB_EVENT_MASK_SUBSTRUCTURE_REDIRECT \
    | XCB_EVENT_MASK_FOCUS_CHANGE \
    | XCB_EVENT_MASK_PROPERTY_CHANGE \
    | XCB_EVENT_MASK_COLOR_MAP_CHANGE \
    | XCB_EVENT_MASK_OWNER_GRAB_BUTTON


static void
create_window(void)
{
	const xcb_setup_t *setup;
	xcb_screen_t *screen;
	uint32_t mask = XCB_CW_BACK_PIXEL | XCB_CW_EVENT_MASK;
	uint32_t values[] = { background_color, ALL_EVENTS };
	static const char name[] = "test3";
	setup = xcb_get_setup(conn);
	if (setup->roots_len < 1)
		panic("no roots");
	screen = xcb_setup_roots_iterator(setup).data;
	win = xcb_generate_id(conn);
	xcb_create_window(
		conn, XCB_COPY_FROM_PARENT, win, screen->root,
		0, 0, 400, 640, 10,
		XCB_WINDOW_CLASS_INPUT_OUTPUT,
		screen->root_visual, mask, values);
	xcb_change_property(conn, XCB_PROP_MODE_REPLACE, win,
		XCB_ATOM_WM_NAME, XCB_ATOM_STRING, 8,
		sizeof(name)-1, name);
	xcb_map_window(conn, win);
	if (xcb_flush(conn) <= 0)
		panic("failed to flush buffer");
}

static void
init(void)
{
	open_connection();
	create_window();
	open_font();
	create_foreground();
}

static void
show_status(void)
{
	static char buf[ARRAY_SIZE(event_counters)][64];
	int i, line_nr = 1, len;
	for (i = 0; i < ARRAY_SIZE(event_counters); i++) {
		if (event_counters[i].name == NULL)
			continue;
		len = snprintf(buf[i], sizeof(buf[i]), "%s: %d",
			event_counters[i].name, event_counters[i].nr);
		xcb_image_text_8(conn, len, win, foreground,
			10, 16 * line_nr, buf[i]);
		line_nr++;
	}
	xcb_flush(conn);
}

static void
handle_event(xcb_generic_event_t *e)
{
	event_counters[e->response_type].nr++;
	show_status();
}

static void
handle_error(xcb_generic_error_t *err)
{
	fprintf(stderr, "got error %d\n", err->error_code);
}

static void
interact(void)
{
	xcb_generic_event_t *e;
	for (;;) {
		e = xcb_wait_for_event(conn);
		if (e == NULL) {
			int err;
			err = xcb_connection_has_error(conn);
			if (err)
				panic(connection_error_message(err));
			panic("unexpected NULL event");
		}
		if (e->response_type)
			handle_event(e);
		else
			handle_error((xcb_generic_error_t *)e);
		free(e);
	}
}

int
main(void)
{
	init();
	interact();
	return 0;
}
